import attr
from django.test import TestCase
from rest_framework import status
from rest_framework.test import APIRequestFactory

from mantle_drf.generics import (
    CreateAPIView,
    DestroyAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from tests.models import Item


@attr.define
class ItemShape:
    name: str
    slug: str


factory = APIRequestFactory()


class TestRetrieveAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="Test", slug="test")

    def test_retrieve(self):
        class TestView(RetrieveAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK
        assert response.data == {"name": "Test", "slug": "test"}


class TestListAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        Item.objects.create(name="A", slug="a")
        Item.objects.create(name="B", slug="b")

    def test_list(self):
        class TestView(ListAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request)
        assert response.status_code == status.HTTP_200_OK
        assert len(response.data) == 2

    def test_list_single_query(self):
        class TestView(ListAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        with self.assertNumQueries(1):
            view(request)

    def test_list_paginated_single_query(self):
        from rest_framework.pagination import PageNumberPagination

        class SmallPagination(PageNumberPagination):
            page_size = 1

        class TestView(ListAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape
            pagination_class = SmallPagination

        request = factory.get("/")
        view = TestView.as_view()
        # 1 query for count, 1 for the page.
        with self.assertNumQueries(2):
            view(request)


class TestCreateAPIView(TestCase):
    def test_create(self):
        class TestView(CreateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.post("/", {"name": "New", "slug": "new"}, format="json")
        view = TestView.as_view()
        response = view(request)
        assert response.status_code == status.HTTP_201_CREATED
        assert response.data["name"] == "New"
        assert Item.objects.filter(slug="new").exists()


class TestDestroyAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="Delete Me", slug="delete-me")

    def test_destroy(self):
        class TestView(DestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.delete("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT
        assert not Item.objects.filter(pk=self.item.pk).exists()


class TestUpdateAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="Old", slug="old")

    def test_update(self):
        class TestView(UpdateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.put(
            "/", {"name": "Updated", "slug": "updated"}, format="json"
        )
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["name"] == "Updated"
        self.item.refresh_from_db()
        assert self.item.name == "Updated"


class TestListCreateAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        Item.objects.create(name="Existing", slug="existing")

    def test_list(self):
        class TestView(ListCreateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request)
        assert response.status_code == status.HTTP_200_OK

    def test_create(self):
        class TestView(ListCreateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.post("/", {"name": "New", "slug": "lc-new"}, format="json")
        view = TestView.as_view()
        response = view(request)
        assert response.status_code == status.HTTP_201_CREATED


class TestRetrieveUpdateAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="RU", slug="ru")

    def test_retrieve(self):
        class TestView(RetrieveUpdateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_update(self):
        class TestView(RetrieveUpdateAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.put(
            "/", {"name": "Updated", "slug": "ru-updated"}, format="json"
        )
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK


class TestRetrieveDestroyAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="RD", slug="rd")

    def test_retrieve(self):
        class TestView(RetrieveDestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_destroy(self):
        class TestView(RetrieveDestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.delete("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT


class TestRetrieveUpdateDestroyAPIView(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="RUD", slug="rud")

    def test_retrieve(self):
        class TestView(RetrieveUpdateDestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.get("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_update(self):
        class TestView(RetrieveUpdateDestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.put(
            "/", {"name": "Updated", "slug": "rud-updated"}, format="json"
        )
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_destroy(self):
        class TestView(RetrieveUpdateDestroyAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        request = factory.delete("/")
        view = TestView.as_view()
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT



from drf_spectacular.utils import extend_schema

from mantle_drf.generics import ListCreateAPIView, RetrieveUpdateDestroyAPIView

from .models import Bookmark
from .shapes import BookmarkShape


class BookmarkList(ListCreateAPIView):
    queryset = Bookmark.objects.all()
    shape_class = BookmarkShape

    @extend_schema(responses=BookmarkShape, tags=["bookmarks"])
    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    @extend_schema(
        request=BookmarkShape, responses=BookmarkShape, tags=["bookmarks"]
    )
    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class BookmarkDetail(RetrieveUpdateDestroyAPIView):
    queryset = Bookmark.objects.all()
    shape_class = BookmarkShape

    @extend_schema(responses=BookmarkShape, tags=["bookmarks"])
    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    @extend_schema(
        request=BookmarkShape, responses=BookmarkShape, tags=["bookmarks"]
    )
    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    @extend_schema(
        request=BookmarkShape, responses=BookmarkShape, tags=["bookmarks"]
    )
    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    @extend_schema(tags=["bookmarks"])
    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)



SECRET_KEY = "example-insecure-key"

DEBUG = True

ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "rest_framework",
    "drf_spectacular",
    "example_app.ExampleAppConfig",
]

ROOT_URLCONF = "urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
    },
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
    }
}

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

REST_FRAMEWORK = {
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
}

SPECTACULAR_SETTINGS = {
    "TITLE": "django-mantle-drf Example",
    "VERSION": "0.1",
}



import attr
import pytest
from django.test import TestCase
from rest_framework.test import APIRequestFactory

from mantle import Query
from mantle_drf.generics import GenericAPIView
from tests.models import Item


@attr.define
class ItemShape:
    name: str
    slug: str


factory = APIRequestFactory()


class TestGetShapeClass(TestCase):
    def test_get_shape_class_returns_shape_class(self):
        class TestView(GenericAPIView):
            shape_class = ItemShape

        view = TestView()
        assert view.get_shape_class() is ItemShape

    def test_get_shape_class_raises_without_shape_class(self):
        class TestView(GenericAPIView):
            pass

        view = TestView()
        with pytest.raises(AssertionError):
            view.get_shape_class()


class TestGetQuery(TestCase):
    def test_get_query_returns_query(self):
        class TestView(GenericAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestView()
        query = view.get_query()
        assert isinstance(query, Query)

    def test_get_query_with_custom_queryset(self):
        class TestView(GenericAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestView()
        qs = Item.objects.filter(name="test")
        query = view.get_query(queryset=qs)
        assert isinstance(query, Query)


class TestStructure(TestCase):
    def test_structure(self):
        class TestView(GenericAPIView):
            shape_class = ItemShape

        view = TestView()
        shape = view.structure({"name": "Test", "slug": "test"})
        assert isinstance(shape, ItemShape)
        assert shape.name == "Test"
        assert shape.slug == "test"


class TestUnstructure(TestCase):
    def test_unstructure(self):
        class TestView(GenericAPIView):
            shape_class = ItemShape

        view = TestView()
        shape = ItemShape(name="Test", slug="test")
        data = view.unstructure(shape)
        assert data == {"name": "Test", "slug": "test"}


class TestGetSerializerClass(TestCase):
    def test_get_serializer_class_returns_none(self):
        """We don't use serializers."""

        class TestView(GenericAPIView):
            shape_class = ItemShape

        view = TestView()
        assert view.get_serializer_class() is None



import attr
from django.test import TestCase
from rest_framework import status
from rest_framework.test import APIRequestFactory

from mantle_drf.viewsets import ModelViewSet, ReadOnlyModelViewSet
from tests.models import Item


@attr.define
class ItemShape:
    name: str
    slug: str


factory = APIRequestFactory()


class TestReadOnlyModelViewSet(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="VS", slug="vs")

    def test_list(self):
        class TestViewSet(ReadOnlyModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"get": "list"})
        request = factory.get("/")
        response = view(request)
        assert response.status_code == status.HTTP_200_OK
        assert len(response.data) == 1

    def test_retrieve(self):
        class TestViewSet(ReadOnlyModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"get": "retrieve"})
        request = factory.get("/")
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK
        assert response.data["name"] == "VS"


class TestModelViewSet(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.item = Item.objects.create(name="CRUD", slug="crud")

    def test_list(self):
        class TestViewSet(ModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"get": "list"})
        request = factory.get("/")
        response = view(request)
        assert response.status_code == status.HTTP_200_OK

    def test_create(self):
        class TestViewSet(ModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"post": "create"})
        request = factory.post("/", {"name": "New", "slug": "vs-new"}, format="json")
        response = view(request)
        assert response.status_code == status.HTTP_201_CREATED

    def test_retrieve(self):
        class TestViewSet(ModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"get": "retrieve"})
        request = factory.get("/")
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_update(self):
        class TestViewSet(ModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"put": "update"})
        request = factory.put(
            "/", {"name": "Updated", "slug": "crud-updated"}, format="json"
        )
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_200_OK

    def test_destroy(self):
        class TestViewSet(ModelViewSet):
            queryset = Item.objects.all()
            shape_class = ItemShape

        view = TestViewSet.as_view({"delete": "destroy"})
        request = factory.delete("/")
        response = view(request, pk=self.item.pk)
        assert response.status_code == status.HTTP_204_NO_CONTENT



SECRET_KEY = "test-secret-key"

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "rest_framework",
    "drf_spectacular",
    "tests",
]

REST_FRAMEWORK = {
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
}

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "tests.urls"

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



from django.apps import AppConfig


class ExampleAppConfig(AppConfig):
    name = "example_app"
    default_auto_field = "django.db.models.BigAutoField"

    def ready(self):
        import mantle_drf.schema  # noqa: F401 — register drf-spectacular extension



"""
drf-spectacular extension for attrs shape classes.
"""
import attr
from drf_spectacular.drainage import set_override
from drf_spectacular.extensions import OpenApiSerializerExtension
from drf_spectacular.plumbing import ResolvedComponent

import jsonschema_extractor


class AttrsExtension(OpenApiSerializerExtension):
    """
    Extension to support attrs classes in @extend_schema(request=..., response=...).
    """

    target_class = "attr.AttrsInstance"

    @classmethod
    def _matches(cls, target) -> bool:
        # attrs protocol isn't runtime-checkable, so use attr.has() instead.
        try:
            return attr.has(target)
        except TypeError:
            return False

    def get_name(self, auto_schema, direction):
        set_override(self.target, "suppress_collision_warning", True)
        return self.target.__name__

    def map_serializer(self, auto_schema, direction):
        schema = jsonschema_extractor.extract(self.target)

        # Register nested definitions as separate OpenAPI components.
        for def_name, def_schema in schema.pop("definitions", {}).items():
            component = ResolvedComponent(
                name=def_name,
                type=ResolvedComponent.SCHEMA,
                object=def_name,
                schema=def_schema,
            )
            auto_schema.registry.register_on_missing(component)

        return schema



"""
Generic views using attrs shape classes instead of DRF serializers.
"""
from mantle import Query
from mantle.db import default_converter
from rest_framework.generics import GenericAPIView as DRFGenericAPIView

from .mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericAPIView(DRFGenericAPIView):
    """
    A DRF GenericAPIView that uses attrs shape classes instead of serializers.

    Set ``shape_class`` on the view to an attrs-decorated class.
    """

    shape_class = None

    def get_shape_class(self):
        """
        Return the shape class to use for structuring/unstructuring.

        Defaults to ``self.shape_class``.
        """
        assert self.shape_class is not None, (
            "'%s' should either include a `shape_class` attribute, "
            "or override the `get_shape_class()` method." % self.__class__.__name__
        )
        return self.shape_class

    def get_query(self, queryset=None):
        """
        Return a ``mantle.Query`` for the current queryset and shape class.
        """
        if queryset is None:
            queryset = self.get_queryset()
        return Query(queryset, self.get_shape_class())

    def structure(self, data):
        """Structure request data into a shape instance."""
        return default_converter.structure(data, self.get_shape_class())

    def unstructure(self, shape):
        """Unstructure a shape instance to a JSON-serialisable dict."""
        return default_converter.unstructure(shape)

    # Override get_serializer_class to return None — we don't use serializers.
    def get_serializer_class(self):
        return None


# Concrete view classes


class CreateAPIView(CreateModelMixin, GenericAPIView):
    """Concrete view for creating a model instance."""

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class ListAPIView(ListModelMixin, GenericAPIView):
    """Concrete view for listing a queryset."""

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(RetrieveModelMixin, GenericAPIView):
    """Concrete view for retrieving a model instance."""

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)


class DestroyAPIView(DestroyModelMixin, GenericAPIView):
    """Concrete view for deleting a model instance."""

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class UpdateAPIView(UpdateModelMixin, GenericAPIView):
    """Concrete view for updating a model instance."""

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(ListModelMixin, CreateModelMixin, GenericAPIView):
    """Concrete view for listing and creating model instances."""

    def get(self, request, *args, **kwargs):
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(RetrieveModelMixin, UpdateModelMixin, GenericAPIView):
    """Concrete view for retrieving and updating a model instance."""

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(RetrieveModelMixin, DestroyModelMixin, GenericAPIView):
    """Concrete view for retrieving and deleting a model instance."""

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(
    RetrieveModelMixin, UpdateModelMixin, DestroyModelMixin, GenericAPIView
):
    """Concrete view for retrieving, updating, and deleting a model instance."""

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.destroy(request, *args, **kwargs)



from django.urls import path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView

from example_app.views import BookmarkDetail, BookmarkList

urlpatterns = [
    path("bookmarks/", BookmarkList.as_view(), name="bookmark-list"),
    path("bookmarks/<int:pk>/", BookmarkDetail.as_view(), name="bookmark-detail"),
    path("schema/", SpectacularAPIView.as_view(), name="schema"),
    path("docs/", SpectacularSwaggerView.as_view(url_name="schema"), name="docs"),
]



from django.db import models


class Bookmark(models.Model):
    url = models.URLField(max_length=500)
    title = models.CharField(max_length=200)
    comment = models.TextField(blank=True, default="")

    def __str__(self):
        return self.title



"""
Mixins providing standard actions for attrs shape-based views.
"""
from rest_framework import status
from rest_framework.response import Response

from mantle import create, update


class CreateModelMixin:
    """Create a model instance from an attrs shape."""

    def create(self, request, *args, **kwargs):
        shape = self.structure(request.data)
        instance = self.perform_create(shape)
        # Re-read via Query to return the persisted state.
        query = self.get_query(
            self.get_queryset().filter(pk=instance.pk)
        )
        data = self.unstructure(query.get())
        return Response(data, status=status.HTTP_201_CREATED)

    def perform_create(self, shape):
        return create(self.get_queryset().model, shape)


class ListModelMixin:
    """List a queryset, returning unstructured shape data."""

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        query = self.get_query(queryset)
        prepared_qs = query.prepare(queryset)

        page = self.paginate_queryset(prepared_qs)
        if page is not None:
            data = [
                self.unstructure(
                    query.converter.structure(query.project(obj), query.shape_class)
                )
                for obj in page
            ]
            return self.get_paginated_response(data)

        data = [
            self.unstructure(
                query.converter.structure(query.project(obj), query.shape_class)
            )
            for obj in prepared_qs
        ]
        return Response(data)


class RetrieveModelMixin:
    """Retrieve a model instance, returning unstructured shape data."""

    def retrieve(self, request, *args, **kwargs):
        instance = self.get_object()
        query = self.get_query(
            self.get_queryset().filter(pk=instance.pk)
        )
        data = self.unstructure(query.get())
        return Response(data)


class UpdateModelMixin:
    """Update a model instance from an attrs shape."""

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        partial = kwargs.pop("partial", False)
        shape = self.structure(request.data)
        instance = self.perform_update(instance, shape, partial=partial)
        query = self.get_query(
            self.get_queryset().filter(pk=instance.pk)
        )
        data = self.unstructure(query.get())
        return Response(data)

    def perform_update(self, instance, shape, partial=False):
        return update(instance, shape)

    def partial_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.update(request, *args, **kwargs)


class DestroyModelMixin:
    """Destroy a model instance."""

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance):
        instance.delete()



import attr
from django.test import TestCase
from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.utils import extend_schema
from rest_framework.response import Response
from rest_framework.test import APIRequestFactory

from mantle_drf.generics import GenericAPIView, RetrieveAPIView
import mantle_drf.schema  # noqa: F401 — registers the extension
from tests.models import Item


@attr.define
class ItemShape:
    name: str
    slug: str


factory = APIRequestFactory()


class TestAttrsExtension(TestCase):
    def _generate_schema(self, urlpatterns):
        generator = SchemaGenerator(patterns=urlpatterns)
        return generator.get_schema(public=True)

    def test_extend_schema_with_attrs_response(self):
        """attrs class used as response generates correct OpenAPI schema."""
        from django.urls import path

        class TestView(RetrieveAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

            @extend_schema(responses=ItemShape)
            def get(self, request, *args, **kwargs):
                return super().get(request, *args, **kwargs)

        urlpatterns = [path("items/<int:pk>/", TestView.as_view())]
        schema = self._generate_schema(urlpatterns)

        item_schema = schema["components"]["schemas"]["ItemShape"]
        assert item_schema["type"] == "object"
        assert "name" in item_schema["properties"]
        assert "slug" in item_schema["properties"]
        assert item_schema["properties"]["name"]["type"] == "string"

    def test_extend_schema_with_attrs_request(self):
        """attrs class used as request generates correct OpenAPI schema."""
        from django.urls import path

        class TestView(GenericAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

            @extend_schema(request=ItemShape, responses=ItemShape)
            def post(self, request, *args, **kwargs):
                return Response({})

        urlpatterns = [path("items/", TestView.as_view())]
        schema = self._generate_schema(urlpatterns)

        # The request body should reference the ItemShape component.
        post_op = schema["paths"]["/items/"]["post"]
        request_body = post_op["requestBody"]
        ref = request_body["content"]["application/json"]["schema"]["$ref"]
        assert "ItemShape" in ref

    def test_schema_required_fields(self):
        """Required fields (no default) appear in 'required' list."""
        from django.urls import path

        class TestView(RetrieveAPIView):
            queryset = Item.objects.all()
            shape_class = ItemShape

            @extend_schema(responses=ItemShape)
            def get(self, request, *args, **kwargs):
                return super().get(request, *args, **kwargs)

        urlpatterns = [path("items/<int:pk>/", TestView.as_view())]
        schema = self._generate_schema(urlpatterns)

        item_schema = schema["components"]["schemas"]["ItemShape"]
        assert "name" in item_schema["required"]
        assert "slug" in item_schema["required"]

    def test_schema_optional_field(self):
        """Optional fields are handled correctly."""
        from typing import Optional
        from django.urls import path

        @attr.define
        class OptionalShape:
            name: str
            label: Optional[str] = None

        class TestView(GenericAPIView):
            queryset = Item.objects.all()
            shape_class = OptionalShape

            @extend_schema(responses=OptionalShape)
            def get(self, request, *args, **kwargs):
                return Response({})

        urlpatterns = [path("items/<int:pk>/", TestView.as_view())]
        schema = self._generate_schema(urlpatterns)

        opt_schema = schema["components"]["schemas"]["OptionalShape"]
        assert "name" in opt_schema["required"]
        # label has a default, so it should not be required.
        assert "label" not in opt_schema.get("required", [])



"""
ViewSets using attrs shape classes instead of DRF serializers.
"""
from rest_framework.viewsets import ViewSetMixin

from .generics import GenericAPIView
from .mixins import (
    CreateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
)


class GenericViewSet(ViewSetMixin, GenericAPIView):
    """
    GenericViewSet with shape class support.

    Does not provide any actions by default.
    """

    pass


class ReadOnlyModelViewSet(RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """ViewSet providing default `list()` and `retrieve()` actions."""

    pass


class ModelViewSet(
    CreateModelMixin,
    RetrieveModelMixin,
    UpdateModelMixin,
    DestroyModelMixin,
    ListModelMixin,
    GenericViewSet,
):
    """ViewSet providing default CRUD actions."""

    pass



from django.db import models


class Item(models.Model):
    name = models.CharField(max_length=100)
    slug = models.SlugField(unique=True)

    class Meta:
        app_label = "tests"
        ordering = ["name"]

    def __str__(self):
        return self.name
